# Overview

* CPU
    * 8-bit processor
    * 1 MHz
    * 16 8-bit registers or 8 16-bit registers
    * Support for memory-mapped I/O and port-mapped I/O
* Memory
    * 16-bit addresses
    * 64 KB (65536 bytes) of addressable memory
    * Memory bank switching
    * DMA transfers
* Video
    * 256x192 pixel screen, 8x8 pixel characters
    * 256 tiles
    * 40 sprites 
    * 16 RGB colors
    * 1-bit color scheme
    * Transparency mask
    * Horizontal and vertical mirroring of tiles and sprites
* Audio
    * 1 channel
    * 4 waveforms: sine, saw, pulse, noise
* Input
    * Two controllers
    * D-pad, SELECT, START, A, B

# Processor

The CPU has 16 8-bit registers or 8 16-bit registers that can all be addressed.

The current implementation of the assembler only supports eight general purpose 8-bit registers.

    A, B, C, D, E, F, G, H

They can be grouped to four 16-bit registers:

    AB, CD, EF, GH

There are three registers that have specific purposes:

    SF - status flags (8-bit)  
    IP - instruction pointer (16-bit)  
    SP - stack pointer (16-bit)

The status flags are set when using certain arithmetic instructions like `ADD` or `CMP`. The status flags are:

    76543210
    ZC------

    Z - zero Flag  
    C - carry Flag

The instruction pointer is the memory address that holds the next instruction. The stack pointer is the memory address of the stack on which you can use `PUSH` and `POP` instructions to store data. It is also used by the processor for the `CALL` and `RET` instruction.

      -------------
    0 |  A  |  B  | 1
      -------------
    2 |  C  |  D  | 3
      -------------
    4 |  E  |  F  | 5
      -------------
    6 |  G  |  H  | 7
      -------------
    8 |  -  |  -  | 9
      -------------
    A |  -  | S R | B
      -------------
    C |    I P    | D
      -------------
    E |    S P    | F
      -------------

The names of the general purpose registers is chosen arbitrarily and have no meaning. You can change them by modifying the assembler code.

## Instructions

The encoding of the instructions follows a very simple design so that they are easy to decode, and easy to extend.

* The OP code is always 8-bit
* Registers are encoded in 8-bit
    * 8-bit registers are indexed from `0x00` to `0x0F`
    * 16-bit registers use the same indices but they are OR'd with the value `0x10`, i.e. the register `CD` is encoded as `0x12`, the regsiter `SP` is encoded as `0x1E`, etc.
* Constants are either 8-bit or 16-bit

This means that an instruction is 32-bit long at most.

For instance, the instruction:

    MOV [ab], $42

Is encoded as:

    14 10 42

    14 - write constant to memory address
    10 - first 16-bit register
    42 - constant

The following letters are used for register names and constants in the instruction listings below:

* `r`  - register (8-bit)  
* `rr` - register (16-bit)  
* `n` - number (8-bit)  
* `nn` - number (16-bit)  
* `[]` - memory access

### System

    NOP         ; no operation instruction
    DI          ; disable interrupts
    EI          ; enable interrupts
    HALT        ; stop the device

### Ports

    IN r,n      ; read data from port N to register 
    IN r,r
    OUT n,r     ; write register value to port N
    OUT r,r

### Memory

    MOV r,n     ; write constant to register
    MOV r,r     ; write register to register
    MOV r,[rr]  ; write data at memory address to register
    MOV [rr],r  ; write register value to memory address
    MOV [rr],n  ; write constant to memory address

    MOV rr,nn
    MOV rr,rr
    MOV [rr],rr ; special case
    MOV rr,[rr] ; special case

    SWP r,r     ; swap register values

    PUSH rr     ; write value to stack (SP--)
    PUSH nn
    POP rr      ; read value from stack (SP++)

### Flow

    JMP nn      ; jump to address
    JC nn       ; jump to address if carry flag is set
    JNC nn      ; jump to address if carry flag is not set
    JZ nn       ; jump to address if zero flag is set
    JNZ nn      ; jump to address if zero flag is not set
    
    CALL nn     ; push IP to stack and jump to address
    RET         ; pop stack to IP
    RETI        ; pop stack to IP and enable interrupts

### Logic

    AND r,n
    AND r,r
    OR r,n
    OR r,r
    XOR r,n
    XOR r,r
    NOT r

### Arithmetic

    ADD r,n     ; addition of two values
    ADD r,r
    ADD rr,nn
    ADD rr,rr
    SUB r,n     ; subtraction of two values
    SUB r,r
    SUB rr,nn
    SUB rr,rr
    INC r       ; increment register value by one
    INC rr
    DEC r       ; decrement register value by one
    DEC rr
    SHL r,n     ; shift register value to the left by N bit
    SHR r,n     ; shift regsiter value to the right by N bit
    
    CMP r,n     ; compare values
    CMP r,r
    CMP rr,nn
    CMP rr,rr

# Memory

The current implementation uses memory-mapped I/O to communicate between devices. The general layout is:

    ADDR        DESCRIPTION
    0000-7FFF   cartridge rom
    8000-9FFF   video memory
    A000-EFFF   ram
    FF00        i/o
    FF80-FFFF   stack

## Cartridge ROM

At the beginning of the cartridge is the interrupt table which should hold addresses to your interrupt handling routines. If you enable the interrupts, the processor will jump to the address given in the interrupt table.

Interrupt table:

    ADDR    DESCRIPTION
    0000    0 VBlank

Banks:

    ADDR    DESCRIPTION
    0000    bank 0
    4000    switchable banks

The ROM bank is switched by writing an 8-bit value to the memory range `$4000` to `$7FFF`. The value must be greater than 0.

For instance, to switch to bank 4:

    mov ab, $4000
    mov [ab], 4

## Video

The screen resolution is 256x192 pixels, 32x24 characters, 768 characters total, 8x8 pixel per character.

    ADDR    DESCRIPTION
    8000    character memory    256 chars, 8 byte per char
    8800    mask memory         256 chars, 8 byte per char
    9000    background memory   768 tiles, 3 bytes per tile
                                    8-bit char index
                                    8-bit fg/bg color
                                    8-bit attribute
                                        1 prio
                                        1 horizontal flip
                                        1 vertical flip
    9900    sprite memory       40 sprites, 5 bytes per sprite
                                    8-bit x position
                                    8-bit y position
                                    8-bit char index
                                    8-bit fg/bg color
                                    8-bit attribute
                                        1 draw
                                        1 horizontal flip
                                        1 vertical flip
    99C8    color palette       16 colors, 3 byte RGB
    99F8    ---

### Character Memory

The character memory holds 256 characters. Each character has a size of 8 byte, 1 byte per row. If the bit is set to 1, the foreground color is used. If the bit is set to 0, the background color is used.

### Mask Memory

The mask memory is applied to the character memory when determining which pixels of a character should be drawn and which not. It is used to render sprites correctly to the background layer.

### Background Memory

The background memory defines what tiles to be drawn to screen. You can define the foreground and background color for each tile on screen. The priority attribute determines if the tile is drawn in front of (1) or behind the sprites (0, default). The tile can be horizontally and vertically mirrored.

### Sprite Memory

The sprite memory holds 40 sprites. Sprites can be moved pixel-per-pixel on the screen. You move the sprite by setting the x and y coordinates. The sprite can be horizontally and vertically mirrored, as well as rotated 90° clock-wise.

## RAM

    ADDR    DESCRIPTION
    A000    bank 0
    C000    switchable banks
    E000    persistent storage (4KB)

## I/O

    ADDR    DESCRIPTION
    FF00    DMA transfer control
	FF07    APU
    FF0A    controller

### DMA

    ADDR    DESCRIPTION
    FF00    source address
    FF02    destination address
    FF04    length in bytes
    FF06    transfer flag

To start the transfer set the flag to `$1`. The flag is reset after the transfer.

### APU

    ADDR    DESCRIPTION
	FF07    waveform: 0 - sine, 1 - saw, 2 - pulse, 3 - noise
	FF08    note (see MIDI table)
	FF09    volume

### Controller

    ADDR    DESCRIPTION
    FF0A    controller 1
    FF0B    controller 2

The state is encoded in one byte:

    7 B
    6 A
    5 SELECT
    4 START
    3 RIGHT
    2 LEFT
    1 DOWN
    0 UP

# Assembler

## Instructions

    .BYTE exp,exp,exp,exp,exp,...
        - writes bytes directly to the binary
    .DEFINE symbol exp
        - defines a symbol that can be used in the assembly code
    .INCBIN "path"
        - includes a binary file directly into the program binary
    .INCLUDE "path"
        - includes a source file
    .ORG addr
        - offsets all addresses in the program by addr
    .ZERO addr
        - fills binary with zeros up to but not including the provided address
