#!/usr/bin/env python

# virt assembler
# 2020-05-16
# Python 3.8.1

import sys

from logging import debug, error

################################################################################
# OPERATION CODES
################################################################################

OP_NOP = 0x00
OP_MOV_R8_N8 = 0x10
OP_MOV_R8_R8 = 0x11
OP_MOV_R8_mR16 = 0x12
OP_MOV_mR16_R8 = 0x13
OP_MOV_mR16_N8 = 0x14
OP_MOV_R16_N16 = 0x15
OP_MOV_R16_R16 = 0x16
OP_MOV_mR16_R16 = 0x17
OP_SWP_R8_R8 = 0x1F
OP_PUSH_R16 = 0x20
OP_PUSH_N16 = 0x21
OP_POP_R16 = 0x22
OP_JMP_N16 = 0x30
OP_JC_N16 = 0x31
OP_JNC_N16 = 0x32
OP_JZ_N16 = 0x33
OP_JNZ_N16 = 0x34
OP_CALL_N16 = 0x40
OP_RET = 0x41
OP_RETI = 0x42
OP_AND_R8_N8 = 0x50
OP_AND_R8_R8 = 0x51
OP_OR_R8_N8 = 0x52
OP_OR_R8_R8 = 0x53
OP_XOR_R8_N8 = 0x54
OP_XOR_R8_R8 = 0x55
OP_NOT_R8 = 0x56
OP_ADD_R8_N8 = 0x60
OP_ADD_R8_R8 = 0x61
OP_ADD_R16_N16 = 0x62
OP_ADD_R16_R16 = 0x63
OP_SUB_R8_N8 = 0x64
OP_SUB_R8_R8 = 0x65
OP_SUB_R16_N16 = 0x66
OP_SUB_R16_R16 = 0x67
OP_INC_R8 = 0x68
OP_INC_R16 = 0x69
OP_DEC_R8 = 0x6A
OP_DEC_R16 = 0x6B
OP_SHL_R8_N8 = 0x6C
OP_SHR_R8_N8 = 0x6D
OP_CMP_R8_N8 = 0x70
OP_CMP_R8_R8 = 0x72
OP_CMP_R16_N16 = 0x73
OP_CMP_R16_R16 = 0x74
OP_IN_R8_N8 = 0x80
OP_IN_R8_R8 = 0x81
OP_OUT_N8_R8 = 0x82
OP_OUT_R8_R8 = 0x83
OP_DI = 0xFD
OP_EI = 0xFE
OP_HALT = 0xFF

FLAG_ZERO = 0x80
FLAG_CARRY = 0x40

def handleNOP(opcode, args):
	data = bytearray()
	data.append(opcode)
	return data
	
def handleN16(opcode, args):
	if args.isdecimal():
		n = int(args)
	else:
		error("Unknown operation: " + str(opcode) + " " + args)
		exit(1)
		
	if not isInt16(n):
		error("Argument is not 16bit")
		exit(1)
	
	data = bytearray()
	data.append(opcode)	
	data.append((n >> 8) & 0xFF)
	data.append(n & 0xFF)
	
	return data

def handleMR16_N8(opcode, args):
	sargs = args.split(",")
	
	if len(sargs) != 2:
		error("Bad syntax, requires 2 arguments")
		exit(1)
	
	data = bytearray()
	
	if isMemoryRegister(sargs[0]) and sargs[1].isdecimal():		
		mr16 = getRegister(sargs[0][1:-1])
		n8 = int(sargs[1])
		
		if not isReg16(sargs[0][1:-1]):
			error("First argument is not 16bit")
			exit(1)
		
		if not isInt8(n8):
			error("Second argument is not 8bit")
			exit(1)
		
		data.append(opcode)
		data.append(mr16)
		data.append(n8)
	else:
		error("Unknown operation: " + str(opcode) + " " + args)
		exit(1)

	return data

def handleMR16_R8(opcode, args):
	sargs = args.split(",")
	
	if len(sargs) != 2:
		error("Bad syntax, requires 2 arguments")
		exit(1)
	
	data = bytearray()
	
	if isMemoryRegister(sargs[0]) and isRegister(sargs[1]):
		if not isReg16(sargs[0][1:-1]):
			error("Memory location register not 16bit")
			exit(1)

		if not isReg8(sargs[1]):
			error("Second argument is not 8bit")
			exit(1)
		
		mr16 = getRegister(sargs[0][1:-1])
		r8 = getRegister(sargs[1])
		
		data.append(opcode)
		data.append(mr16)
		data.append(r8)
	else:
		error("Unknown operation: " + str(opcode) + " " + args)
		exit(1)

	return data

def handleMR16_R16(opcode, args):
	sargs = args.split(",")
	
	if len(sargs) != 2:
		error("Bad syntax, requires 2 arguments")
		exit(1)
	
	data = bytearray()
	
	if isMemoryRegister(sargs[0]) and isRegister(sargs[1]):
		if not isReg16(sargs[0][1:-1]):
			error("Memory location register not 16bit")
			exit(1)
		
		if not isReg16(sargs[1]):
			error("Second argument is not 16bit")
			exit(1)
		
		mr16 = getRegister(sargs[0][1:-1])
		r16 = getRegister(sargs[1])
		
		data.append(opcode)
		data.append(mr16)
		data.append(r16)
	else:
		error("Unknown operation: " + str(opcode) + " " + args)
		exit(1)
	
	return data

def handleReadR8(opcode, args):
	data = bytearray()

	if isRegister(args):
		if not isReg8(args):
			error("Register  is not 8 bit")
			exit(1)
		
		r = getRegister(args)
		
		data.append(opcode)
		data.append(r)
	else:
		error("Unknown operation: " + str(opcode) + " " + args)
		exit(1)
	
	return data

def handleWriteR8(opcode, args):
	data = bytearray()

	if isRegister(args):
		if not isReg8(args):
			error("Register  is not 8 bit")
			exit(1)
				
		r = getRegister(args)
		
		data.append(opcode)
		data.append(r)
	else:
		error("Unknown operation: " + str(opcode) + " " + args)
		exit(1)
	
	return data

def handleR8_mR16(opcode, args):
	sargs = args.split(",")
	
	if len(sargs) != 2:
		error("Bad syntax, requires 2 arguments")
		exit(1)
	
	data = bytearray()
	
	if isRegister(sargs[0]) and isMemoryRegister(sargs[1]):				
		if not isReg8(sargs[0]):
			error("First register is not 8 bit")
			exit(1)
				
		if not isReg16(sargs[1][1:-1]):
			error("Memory location register not 16 bit")
			exit(1)
		
		r8 = getRegister(sargs[0])
		mr16 = getRegister(sargs[1][1:-1])
		
		data.append(opcode)
		data.append(r8)
		data.append(mr16)
	else:
		error("Unknown operation: " + str(opcode) + " " + args)
		exit(1)

	return data

def handleR8_N8(opcode, args):
	sargs = args.split(",")
	
	if len(sargs) != 2:
		error("Bad syntax, requires 2 arguments")
		exit(1)
	
	data = bytearray()
	
	if isRegister(sargs[0]) and sargs[1].isdecimal():
		r = getRegister(sargs[0])
		n = int(sargs[1])
		
		if not isReg8(sargs[0]):
			error("First register is not 8 bit")
			exit(1)
				
		if not isInt8(n):
			error("Second argument is not 8 bit")
			exit(1)
		
		data.append(opcode)
		data.append(r)
		data.append(n)
	else:
		error("Unknown operation: " + str(opcode) + " " + args)
		exit(1)
	
	return data

def handleN8_R8(opcode, args):
	sargs = args.split(",")
	
	if len(sargs) != 2:
		error("Bad syntax, requires 2 arguments")
		exit(1)
	
	data = bytearray()
	
	if sargs[0].isdecimal() and isRegister(sargs[1]):
		n = int(sargs[0])
		r = getRegister(sargs[1])
		
		if not isInt8(n):
			error("First argument is not 8 bit")
			exit(1)

		if not isReg8(sargs[1]):
			error("Second register is not 8 bit")
			exit(1)
		
		data.append(opcode)
		data.append(n)
		data.append(r)
	else:
		error("Unknown operation: " + str(opcode) + " " + args)
		exit(1)
	
	return data

def handleR8_R8(opcode, args):
	sargs = args.split(",")
	
	if len(sargs) != 2:
		error("Bad syntax, requires 2 arguments")
		exit(1)
	
	data = bytearray()
	
	if isRegister(sargs[0]) and isRegister(sargs[1]):		
		if not isReg8(sargs[0]):
			error("First register is not 8 bit")
			exit(1)
				
		if not isReg8(sargs[1]):
			error("Second register is not 8 bit")
			exit(1)
		
		r1 = getRegister(sargs[0])
		r2 = getRegister(sargs[1])
		
		data.append(opcode)
		data.append(r1)
		data.append(r2)
	else:
		error("Unknown operation: " + str(opcode) + " " + args)
		exit(1)
	
	return data

def handleReadR16(opcode, args):
	data = bytearray()

	if isRegister(args):
		if not isReg16(args):
			error("Register is not 16 bit")
			exit(1)
				
		r = getRegister(args)
		
		data.append(opcode)
		data.append(r)
	else:
		error("Unknown operation: " + str(opcode) + " " + args)
		exit(1)

	return data
	
def handleWriteR16(opcode, args):
	data = bytearray()

	if isRegister(args):
		if not isReg16(args):
			error("Register is not 16 bit")
			exit(1)
				
		r = getRegister(args)
		
		data.append(opcode)
		data.append(r)
	else:
		error("Unknown operation: " + str(opcode) + " " + args)
		exit(1)

	return data

def handleR16_N16(opcode, args):
	sargs = args.split(",")
	
	if len(sargs) != 2:
		error("Bad syntax, requires 2 arguments")
		exit(1)
	
	data = bytearray()
	
	if isRegister(sargs[0]) and sargs[1].isdecimal():
		r = getRegister(sargs[0])
		n = int(sargs[1])
		
		if not isReg16(sargs[0]):
			error("Register is not 16 bit")
			exit(1)
				
		if not isInt16(n):
			error("Second argument is not 16 bit")
			exit(1)
		
		data.append(opcode)
		data.append(r)
		data.append((n >> 8) & 0xFF)
		data.append(n & 0xFF)
	else:
		error("Unknown operation: " + str(opcode) + " " + args)
		exit(1)
	
	return data
	
def handleR16_R16(opcode, args):
	sargs = args.split(",")
	
	if len(sargs) != 2:
		error("Bad syntax, requires 2 arguments")
		exit(1)
	
	data = bytearray()
	
	if isRegister(sargs[0]) and isRegister(sargs[1]):		
		if not isReg16(sargs[0]):
			error("First register is not 16 bit")
			exit(1)
				
		if not isReg16(sargs[1]):
			error("Second register is not 16 bit")
			exit(1)
		
		r1 = getRegister(sargs[0])
		r2 = getRegister(sargs[1])
		
		data.append(opcode)
		data.append(r1)
		data.append(r2)
	else:
		error("Unknown operation: " + str(opcode) + " " + args)
		exit(1)
	
	return data

################################################################################
# ASSEMBLER INSTRUCTIONS
################################################################################

def handleByte(args):
	sargs = args.split(",")
	
	data = bytearray()
	
	for exp in sargs:
		if isNumber(exp):
			n = getNumber(exp)

			if isInt8(n):	
				data.append(n)
			elif isInt16(n):
				data.append((n >> 8) & 0xFF)
				data.append(n & 0xFF)
			else:
				error("Number is not 8 or 16 bit")
				exit(1)
		elif isString(exp):
			for s in list(exp[1:-1]):
				n = ord(s)
				assert isInt8(n), chr(n) + " in " + exp

				data.append(n)
	
	return data	

def handleDefine(args):
	sargs = args.partition(" ")
	
	if isSymbol(sargs[0]) and isNumber(sargs[2]):
		if sargs[0] in addressMap:
			error("Symbol already exists " + sargs[0])
			exit(1)
	
		n = getNumber(sargs[2])
			
		if isInt16(n):
			addressMap[sargs[0]] = getNumber(sargs[2])
		else:
			error("Must be 16bit: " + sargs[2])
			exit(1)
	elif isRegister(sargs[0]):
		error("Cannot use register name as symbol (case of letters is ignored)\n")
		exit(1)
	else:
		error("Bad format " + sargs[0])
		exit(1)	
	
	return bytearray()

def handleIncbin(args):
	if not isString(args):
		error("Not a string: " + args)
		return bytearray()
		
	f = open(args[1:-1], "rb")
	data = bytearray(f.read())
	f.close()
	
	return data

def handleOrg(args):
	global addressPtr
	
	decimal = getNumber(args)
	
	if isInt16(decimal):
		addressPtr = decimal
	else:
		error("ORG address must be 16 bit")
		exit(1)
	
	return bytearray()

def handleZero(args):
	global addressPtr

	# function has to be able to handle addresses >16bit
	addr = getNumber(args)	
	assert addr > addressPtr

	diff = addr - addressPtr
	data = bytearray()

	while diff > 0:
		data.append(0)
		diff -= 1
	
	return data	

################################################################################

def handleInstruction(line):
	global addressPtr
	
	if asmPass == 2 and line.startswith(".define"):
		return bytearray()
	
	line = line[1:]
	part = line.partition(" ")
	op = part[0]
	args = part[2]

	ff = instructions[op]
	data = ff(args)
	
	addressPtr += len(data)
	
	return data

def handleOperation(line):
	global addressPtr
	
	part = line.partition(" ")
	op = part[0]
	args = part[2]
		
	opcode = getOpCode(line)
		
	ff = opdict[opcode][1]
	data = ff(opcode, args)
	
	addressPtr += len(data)
	
	return data

################################################################################
# UTILS
################################################################################

def getNumber(exp):
	n = -1
		
	if isBinary(exp):
		n = int(exp[1:],2)
	elif isDecimal(exp):
		n = int(exp)
	elif isHex(exp):
		n = int(exp[1:], 16)
	else:
		error("Unknown type: " + exp)
	
	return n

def getOpCode(line):
	if not isOperation(line):
		error("Not an operation")
		exit(1)

	part = line.partition(" ")
	op = part[0]
	args = part[2]
	sargs = args.split(",")
	
	opcode = -1
	
	if op == "add" and len(sargs) == 2:
		if isReg8(sargs[0]) and isExpression(sargs[1]):
			opcode = OP_ADD_R8_N8
		elif isReg8(sargs[0]) and isReg8(sargs[1]):
			opcode = OP_ADD_R8_R8
		elif isReg16(sargs[0]) and isExpression(sargs[1]):
			opcode = OP_ADD_R16_N16
		elif isReg16(sargs[0]) and isReg16(sargs[1]):
			opcode = OP_ADD_R16_R16
	elif op == "and" and len(sargs) == 2:
		if isReg8(sargs[0]) and isExpression(sargs[1]):
			opcode = OP_AND_R8_N8
		elif isReg8(sargs[0]) and isReg8(sargs[1]):
			opcode = OP_AND_R8_R8
	elif op == "call":
		if isExpression(sargs[0]):
			opcode = OP_CALL_N16
	elif op == "cmp" and len(sargs) == 2:
		if isReg8(sargs[0]) and isExpression(sargs[1]):
			opcode = OP_CMP_R8_N8
		elif isReg8(sargs[0]) and isReg8(sargs[1]):
			opcode = OP_CMP_R8_R8
		elif isReg16(sargs[0]) and isExpression(sargs[1]):
			opcode = OP_CMP_R16_N16
		elif isReg16(sargs[0]) and isReg16(sargs[1]):
			opcode = OP_CMP_R16_R16
	elif op == "dec":
		if isReg8(sargs[0]):
			opcode = OP_DEC_R8
		elif isReg16(sargs[0]):
			opcode = OP_DEC_R16
	elif op == "di":
		opcode = OP_DI
	elif op == "ei":
		opcode = OP_EI
	elif op == "halt":
		opcode = OP_HALT
	elif op == "in":
		if isReg8(sargs[0]) and isExpression(sargs[1]):
			opcode = OP_IN_R8_N8
		elif isReg8(sargs[0]) and isReg8(sargs[1]):
			opcode = OP_IN_R8_R8
	elif op == "inc":
		if isReg8(sargs[0]):
			opcode = OP_INC_R8
		elif isReg16(sargs[0]):
			opcode = OP_INC_R16
	elif op == "jmp":
		if isExpression(sargs[0]):
			opcode = OP_JMP_N16
	elif op == "jnc":
		if isExpression(sargs[0]):
			opcode = OP_JNC_N16
	elif op == "jc":
		if isExpression(sargs[0]):
			opcode = OP_JC_N16
	elif op == "jnz":
		if isExpression(sargs[0]):
			opcode = OP_JNZ_N16
	elif op == "jz":
		if isExpression(sargs[0]):
			opcode = OP_JZ_N16
	elif op == "mov" and len(sargs) == 2:
		if isReg8(sargs[0]) and isExpression(sargs[1]):
			opcode = OP_MOV_R8_N8
		elif isReg8(sargs[0]) and isReg8(sargs[1]):
			opcode = OP_MOV_R8_R8
		elif isReg8(sargs[0]) and isMemoryRegister(sargs[1]):
			opcode = OP_MOV_R8_mR16
		elif isMemoryRegister(sargs[0]) and isReg8(sargs[1]):
			opcode = OP_MOV_mR16_R8
		elif isMemoryRegister(sargs[0]) and isReg16(sargs[1]):
			opcode = OP_MOV_mR16_R16
		elif isMemoryRegister(sargs[0]) and isExpression(sargs[1]):
			opcode = OP_MOV_mR16_N8
		elif isReg16(sargs[0]) and isExpression(sargs[1]):
			opcode = OP_MOV_R16_N16
		elif isReg16(sargs[0]) and isReg16(sargs[1]):
			opcode = OP_MOV_R16_R16
	elif op == "not":
		if isReg8(sargs[0]):
			opcode = OP_NOT_R8
	elif op == "nop":
		opcode = OP_NOP
	elif op == "or" and len(sargs) == 2:
		if isReg8(sargs[0]) and isExpression(sargs[1]):
			opcode = OP_OR_R8_N8
		elif isReg8(sargs[0]) and isReg8(sargs[1]):
			opcode = OP_OR_R8_R8
	elif op == "out" and len(sargs) == 2:
		if isExpression(sargs[0]) and isReg8(sargs[1]):
			opcode = OP_OUT_N8_R8
		elif isReg8(sargs[0]) and isReg8(sargs[1]):
			opcode = OP_OUT_R8_R8
	elif op == "pop":
		if isReg16(sargs[0]):
			opcode = OP_POP_R16
	elif op == "push":
		if isReg16(sargs[0]):
			opcode = OP_PUSH_R16
		elif isExpression(sargs[0]):
			opcode = OP_PUSH_N16
	elif op == "ret":
		opcode = OP_RET
	elif op == "reti":
		opcode = OP_RETI
	elif op == "shl":
		if isReg8(sargs[0]) and isExpression(sargs[1]):
			opcode = OP_SHL_R8_N8
	elif op == "shr":
		if isReg8(sargs[0]) and isExpression(sargs[1]):
			opcode = OP_SHR_R8_N8
	elif op == "sub" and len(sargs) == 2:
		if isReg8(sargs[0]) and isExpression(sargs[1]):
			opcode = OP_SUB_R8_N8
		elif isReg8(sargs[0]) and isReg8(sargs[1]):
			opcode = OP_SUB_R8_R8
		elif isReg16(sargs[0]) and isExpression(sargs[1]):
			opcode = OP_SUB_R16_N16
		elif isReg16(sargs[0]) and isReg16(sargs[1]):
			opcode = OP_SUB_R16_R16
	elif op == "swp" and len(sargs) == 2:
		if isReg8(sargs[0]) and isReg8(sargs[1]):
			opcode = OP_SWP_R8_R8
	elif op == "xor" and len(sargs) == 2:
		if isReg8(sargs[0]) and isExpression(sargs[1]):
			opcode = OP_XOR_R8_N8
		elif isReg8(sargs[0]) and isReg8(sargs[1]):
			opcode = OP_XOR_R8_R8
	
	if opcode == -1:
		error("getOpCode(): Unknown operation: " + line + " " + str(opcode))
		exit(1)
	
	return opcode
	
def getRegister(exp):
	i = regnames.index(exp)
	return regcode[i]

def isBinary(exp):
	chars = "01"
	return exp.startswith("%") and all(c in chars for c in exp[1:])

def isDecimal(exp):
	return exp.isdecimal()

def isExpression(exp):
	return isNumber(exp) or isSymbol(exp)
	
def isHex(exp):
	hexdigits = "0123456789abcdefABCDEF"
	return exp.startswith("$") and all(c in hexdigits for c in exp[1:])
	
def isInstruction(line):
	line = line.lower()

	for i in instructions:
		if line[1:].find(i + " ") == 0:
			return True
	
	return False

def isInt8(i):
	return 0x0 <= i and i <= 0xFF

def isReg8(r):
	if r not in regnames:
		return False

	r = r.lower()
	i = regnames.index(r)
	return regsizes[i] == 8
	
def isReg16(r):
	if r not in regnames:
		return False

	r = r.lower()
	i = regnames.index(r)
	return regsizes[i] == 16

def isInt16(i):
	return 0x0 <= i and i <= 0xFFFF
	
def isMemoryAddress(exp):
	return exp.startswith("[") and exp.endswith("]") and isExpression(exp[1:-1])

def isMemoryRegister(exp):
	return exp.startswith("[") and exp.endswith("]") and isRegister(exp[1:-1])

def isNumber(exp):
	return isBinary(exp) or isDecimal(exp) or isHex(exp)

def isLabel(line):
	return line.endswith(":") and isSymbol(line[:line.find(":")])
	
def isOperation(line):
	line = line.lower()

	for o in operations:
		# i.e. check for "push " or "nop"
		if line.find(o + " ") == 0 or (line.find(o) == 0 and len(line) == len(o)):
			return True
	
	return False
	
def isRegister(exp):
	exp = exp.lower()
	return exp in regnames

def isString(exp):
	return exp[0] == '"' and exp[-1] == '"'

def isSymbol(exp):
	if isNumber(exp) or isRegister(exp):
		return False

	chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_"
	return exp not in regnames and all(c in chars for c in exp)

def resolveSymbol(symbol):
	if not isSymbol(symbol):
		return -1
	
	if symbol not in addressMap:
		error("Cannot resolve symbol: " + symbol)
		exit(1)

	return int(addressMap.get(symbol))
	
################################################################################
	
def normalize(line):
	# remove comments
	if line.startswith(";"):
		return ""
	
	if line.find(";") != -1:
		line = line[:line.find(";")]
	
	# normalize
	line = line.strip()
	
	if len(line) == 0:
		return line
	
	line = line.replace("\t", " ")

	# replace all access spaces
	while line.find("  ") != -1:
		line = line.replace("  ", " ")
	
	if isOperation(line) or isInstruction(line):
		while line.find(", ") != -1:
			line = line.replace(", ", ",")
		
		while line.find(" ,") != -1:
			line = line.replace(" ,", ",")
	
	# lower register letters and turn any number to decimal
	if isOperation(line):
		if " " in line:
			part = line.partition(" ")
			op = part[0].lower()
			args = part[2]
			
			sargs = args.split(",")
			
			line = op + " "
			
			for i in range(len(sargs)):
				if i > 0:
					line += ","

				if isNumber(sargs[i]):
					line += str(getNumber(sargs[i]))
				elif isRegister(sargs[i]):
					line += str(sargs[i].lower())
				elif isMemoryAddress(sargs[i]) and isNumber(sargs[i][1:-1]):
					line += "[" + str(getNumber(sargs[i][1:-1])) + "]"
				else:
					line += sargs[i]
		else:
			line = line.lower()
	
	if isInstruction(line):
		if " " in line:
			part = line.partition(" ")
			
			line = part[0].lower() + " " + part[2]
		else:
			line = line.lower()
		
	return line
	
def resolveSymbols(line):
	if isOperation(line):
		if " " in line:
			part = line.partition(" ")
			op = part[0].lower()
			args = part[2]
			
			sargs = args.split(",")
			
			line = op + " "
			
			for i in range(len(sargs)):
				if i > 0:
					line += ","

				if isSymbol(sargs[i]):
					addr = resolveSymbol(sargs[i])
					
					if addr != -1:
						line += str(addr)
					else:
						error("Symbol cannot be resolved: " + sargs[i])
						exit(1)
				elif isMemoryAddress(sargs[i]) and isSymbol(sargs[i][1:-1]):
					addr = resolveSymbol(sargs[i][1:-1])
				
					if addr != -1:
						line += "[" + str(addr) + "]"
					else:
						error("Symbol cannot be resolved: " + sargs[i])
						exit(1)
				else:
					line += sargs[i]
	return line

def expand(src):
	lines = []
	with open(src, "r") as file:
		for line in file:
			line = normalize(line)
			
			if len(line) == 0:
				continue

			if line.startswith(".include"):
				part = line.partition(" ")
				args = part[2]

				lines.extend(expand(args[1:-1]))
			else:
				lines.append(line)

	return lines

################################################################################
# ASSEMBLER
################################################################################

### First pass: create symbol table ############################################

def resolve(lines):
	global addressPtr, asmLine
	
	debug("--- Resolve ---")
	
	for line in lines:
		asmLine = asmLine + 1

		if isOperation(line):
			opcode = getOpCode(line)
			
			if opcode != -1:
				debug(str(asmLine) + ": " + hex(addressPtr) + " " + str(opdict[opcode][0]) + " " + hex(opcode) + " " + line)
				addressPtr += opdict[opcode][0]
			else:
				error("Cannot find opcode for operation: " + line)
				exit(1)
		elif isInstruction(line):
			handleInstruction(line)
			debug(line)
		elif isLabel(line):
			label = line[:-1]

			if label not in addressMap:
				addressMap[label] = addressPtr
			else:
				error("Label already exists " + label)
				exit(1)
		else:
			error("Unknown operation: " + line)
			exit(1)

	newlines = []
	for line in lines:
		if isLabel(line):
			continue
		elif isInstruction(line) and not (line.startswith(".byte")
				or line.startswith(".incbin")
				or line.startswith(".zero")):
			continue

		newlines.append(resolveSymbols(line))

	return newlines

### Second pass: assemble instructions #########################################

def assemble(lines, dst):
	global addressPtr, asmLine, fbin

	debug("--- Assemble ---")

	fbin = open(dst, "wb")

	for line in lines:
		asmLine = asmLine + 1
		line = line.strip()

		debug(str(asmLine) + " " + str(addressPtr) + ": " + line)

		if isOperation(line):
			fbin.write(handleOperation(line))
		elif isInstruction(line):
			fbin.write(handleInstruction(line))

	fbin.close()

def asm(src, dst):
	global addressPtr, asmLine, asmPass

	lines = expand(src)

	addressPtr = 0
	asmLine = 0
	asmPass = 1
	lines = resolve(lines)
	
	addressPtr = 0
	asmLine = 0
	asmPass = 2
	assemble(lines, dst)

### GLOBALS ####################################################################

addressPtr = 0
asmLine = 0
asmPass = 0
fbin = 0

addressMap = {}

regcode  = [ 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x10, 0x12, 0x14, 0x16, 0x1E ]
regsizes = [    8,    8,    8,    8,    8,    8,    8,    8,   16,   16,   16,   16,   16 ]
regnames = [  "a",  "b",  "c",  "d",  "e",  "f",  "g",  "h", "ab", "cd", "ef", "gh", "sp" ]

instructions = {
	"define": handleDefine,
	"include": None,
	"org": handleOrg,
	#
	"byte": handleByte,
	"incbin": handleIncbin,
	"zero": handleZero
} # check exceptions in resolve

operations = [
	"add", "and", "call", "cmp", "dec", "di", "ei", "halt", "in", "inc", "jmp",
	"jc", "jnc", "jnz", "jz", "mov", "nop", "not", "or", "out", "pop", "push",
	"ret", "reti", "shl", "shr", "sub", "swp", "xor"
]

opdict = {
	OP_NOP:          [1, handleNOP],
	OP_DI:           [1, handleNOP],
	OP_EI:           [1, handleNOP],
	OP_HALT:         [1, handleNOP],
	OP_MOV_R8_N8:    [3, handleR8_N8],
	OP_MOV_R8_R8:    [3, handleR8_R8],
	OP_MOV_R8_mR16:  [3, handleR8_mR16],
	OP_MOV_mR16_R8:  [3, handleMR16_R8],
	OP_MOV_mR16_N8:  [3, handleMR16_N8],
	OP_MOV_R16_N16:  [4, handleR16_N16],
	OP_MOV_R16_R16:  [3, handleR16_R16],
	OP_MOV_mR16_R16: [3, handleMR16_R16],
	OP_SWP_R8_R8:    [3, handleR8_R8],
	OP_PUSH_R16:     [2, handleReadR16],
	OP_PUSH_N16:     [3, handleN16],
	OP_POP_R16:      [2, handleWriteR16],
	OP_JMP_N16:      [3, handleN16],
	OP_JC_N16:       [3, handleN16],
	OP_JNC_N16:      [3, handleN16],
	OP_JZ_N16:       [3, handleN16],
	OP_JNZ_N16:      [3, handleN16],
	OP_CALL_N16:     [3, handleN16],
	OP_RET:          [1, handleNOP],
	OP_RETI:         [1, handleNOP],
	OP_AND_R8_N8:    [3, handleR8_N8],
	OP_AND_R8_R8:    [3, handleR8_R8],
	OP_OR_R8_N8:     [3, handleR8_N8],
	OP_OR_R8_R8:     [3, handleR8_R8],
	OP_XOR_R8_N8:    [3, handleR8_N8],
	OP_XOR_R8_R8:    [3, handleR8_R8],
	OP_NOT_R8:       [2, handleWriteR8],
	OP_ADD_R8_N8:    [3, handleR8_N8],
	OP_ADD_R8_R8:    [3, handleR8_R8],
	OP_ADD_R16_N16:  [4, handleR16_N16],
	OP_ADD_R16_R16:  [3, handleR16_R16],
	OP_SUB_R8_N8:    [3, handleR8_N8],
	OP_SUB_R8_R8:    [3, handleR8_R8],
	OP_SUB_R16_N16:  [4, handleR16_N16],
	OP_SUB_R16_R16:  [3, handleR16_R16],
	OP_INC_R8:       [2, handleWriteR8],
	OP_INC_R16:      [2, handleWriteR16],
	OP_DEC_R8:       [2, handleWriteR8],
	OP_DEC_R16:      [2, handleWriteR16],
	OP_SHL_R8_N8:    [3, handleR8_N8],
	OP_SHR_R8_N8:    [3, handleR8_N8],
	OP_CMP_R8_N8:    [3, handleR8_N8],
	OP_CMP_R8_R8:    [3, handleR8_R8],
	OP_CMP_R16_N16:  [4, handleR16_N16],
	OP_CMP_R16_R16:  [3, handleR16_R16],
	OP_IN_R8_N8:     [3, handleR8_N8], 
	OP_IN_R8_R8:     [3, handleR8_R8],
	OP_OUT_N8_R8:    [3, handleN8_R8],
	OP_OUT_R8_R8:    [3, handleR8_R8],
}

def main():
	global fbin
	
	src = sys.argv[1]
	dst = sys.argv[2]
	
	asm(src, dst)

	print("\n--- Address Map ---")

	for s in addressMap:
		print(str(hex(addressMap.get(s))) + "\t" + s)
			
if __name__ == "__main__":
	main()
